@@ -13,6 +13,9 @@ use super::rewrite::CaveatError;
 use super::rewrite::CheckedCaveat;
 use super::schemas::sturdy;
 
+use parking_lot::Mutex;
+use parking_lot::RwLock;
+
 use preserves::value::ArcValue;
 use preserves::value::Domain;
 use preserves::value::IOValue;
@@ -31,8 +34,6 @@ use std::convert::TryInto;
 use std::marker::PhantomData;
 use std::num::NonZeroU64;
 use std::sync::Arc;
-use std::sync::Mutex;
-use std::sync::RwLock;
 use std::sync::Weak;
 use std::sync::atomic::{AtomicI64, AtomicU64, Ordering};
 use std::time;
@@ -355,8 +356,8 @@ pub struct RunningActor {
 /// values. Use [`Activation::dataflow`] to create a reactive block within a facet that will be
 /// (re-)executed whenever some dependent field changes value.
 ///
-#[derive(Debug)]
 pub struct Field<T: Any + Send> {
+    pub name: String,
     pub field_id: FieldId,
     tx: UnboundedSender<SystemMessage>,
     phantom: PhantomData<T>,
@@ -453,6 +454,16 @@ pub enum RunDisposition {
     Terminate(ActorResult),
 }
 
+/// [Linked tasks][Activation::linked_task] terminate yielding values of this type.
+pub enum LinkedTaskTermination {
+    /// Causes the task's associated [Facet] to be [stop][Activation::stop]ped when the task
+    /// returns.
+    Normal,
+    /// Causes no action to be taken regarding the task's associated [Facet] at the time the
+    /// task returns.
+    KeepFacet,
+}
+
 //---------------------------------------------------------------------------
 
 const BUMP_AMOUNT: u8 = 10;
@@ -495,8 +506,10 @@ preserves_schema::support::lazy_static! {
     };
 
     #[doc(hidden)]
-    pub static ref ACCOUNTS: RwLock<Map<u64, (tracing::Span, Arc<AtomicI64>)>> =
-        RwLock::new(Map::new());
+    pub static ref ACCOUNTS: RwLock<Map<u64, (tracing::Span, Arc<AtomicI64>)>> = Default::default();
+
+    #[doc(hidden)]
+    pub static ref ACTORS: RwLock<Map<ActorId, (tracing::Span, ActorRef)>> = Default::default();
 }
 
 impl TryFrom<&AnyValue> for Synced {
@@ -566,49 +579,47 @@ impl FacetRef {
     ) -> ActorResult where
         F: FnOnce(&mut Activation) -> RunDisposition,
     {
-        match self.actor.state.lock() {
-            Err(_) => panicked_err(),
-            Ok(mut g) => match &mut *g {
-                ActorState::Terminated { exit_status } =>
-                    Err(error("Could not activate terminated actor",
-                              encode_error((**exit_status).clone()))),
-                ActorState::Running(state) => {
-                    tracing::trace!(actor_id=?self.actor.actor_id, "activate");
-                    let mut activation = Activation::make(self, account, state);
-                    let f_result = f(&mut activation);
-                    let result = match activation.restore_invariants(f_result) {
-                        RunDisposition::Continue => Ok(()),
-                        RunDisposition::Terminate(exit_status) => {
-                            if exit_status.is_err() {
-                                activation.clear();
-                            }
-                            drop(activation);
-                            let exit_status = Arc::new(exit_status);
-                            let mut t = Activation::make(&self.actor.facet_ref(state.root),
-                                                         Account::new(crate::name!("shutdown")),
-                                                         state);
-                            if let Err(err) = t._terminate_facet(t.state.root, exit_status.is_ok()) {
-                                // This can only occur as the result of an internal error in this file's code.
-                                tracing::error!(?err, "unexpected error from terminate_facet");
-                                panic!("Unexpected error result from terminate_facet");
-                            }
-                            // TODO: The linked_tasks are being cancelled above ^ when their Facets drop.
-                            // TODO: We don't want that: we want (? do we?) exit hooks to run before linked_tasks are cancelled.
-                            // TODO: Example: send an error message in an exit_hook that is processed and delivered by a linked_task.
-                            for action in std::mem::take(&mut t.state.exit_hooks) {
-                                if let Err(err) = action(&mut t, &exit_status) {
-                                    tracing::error!(?err, "error in exit hook");
-                                }
-                            }
-                            *g = ActorState::Terminated {
-                                exit_status: Arc::clone(&exit_status),
-                            };
-                            (*exit_status).clone()
+        let mut g = self.actor.state.lock();
+        match &mut *g {
+            ActorState::Terminated { exit_status } =>
+                Err(error("Could not activate terminated actor",
+                          encode_error((**exit_status).clone()))),
+            ActorState::Running(state) => {
+                tracing::trace!(actor_id=?self.actor.actor_id, "activate");
+                let mut activation = Activation::make(self, account, state);
+                let f_result = f(&mut activation);
+                let result = match activation.restore_invariants(f_result) {
+                    RunDisposition::Continue => Ok(()),
+                    RunDisposition::Terminate(exit_status) => {
+                        if exit_status.is_err() {
+                            activation.clear();
                         }
-                    };
-                    tracing::trace!(actor_id=?self.actor.actor_id, "deactivate");
-                    result
-                }
+                        drop(activation);
+                        let exit_status = Arc::new(exit_status);
+                        let mut t = Activation::make(&self.actor.facet_ref(state.root),
+                                                     Account::new(crate::name!("shutdown")),
+                                                     state);
+                        if let Err(err) = t._terminate_facet(t.state.root, exit_status.is_ok()) {
+                            // This can only occur as the result of an internal error in this file's code.
+                            tracing::error!(?err, "unexpected error from terminate_facet");
+                            panic!("Unexpected error result from terminate_facet");
+                        }
+                        // TODO: The linked_tasks are being cancelled above ^ when their Facets drop.
+                        // TODO: We don't want that: we want (? do we?) exit hooks to run before linked_tasks are cancelled.
+                        // TODO: Example: send an error message in an exit_hook that is processed and delivered by a linked_task.
+                        for action in std::mem::take(&mut t.state.exit_hooks) {
+                            if let Err(err) = action(&mut t, &exit_status) {
+                                tracing::error!(?err, "error in exit hook");
+                            }
+                        }
+                        *g = ActorState::Terminated {
+                            exit_status: Arc::clone(&exit_status),
+                        };
+                        (*exit_status).clone()
+                    }
+                };
+                tracing::trace!(actor_id=?self.actor.actor_id, "deactivate");
+                result
             }
         }
     }
@@ -874,9 +885,9 @@ impl<'activation> Activation<'activation> {
 
     /// Start a new [linked task][crate::actor#linked-tasks] attached to the active facet. The
     /// task will execute the future "`boot`" to completion unless it is cancelled first (by
-    /// e.g. termination of the owning facet or crashing of the owning actor). Uses `name` for
-    /// log messages emitted by the task.
-    pub fn linked_task<F: 'static + Send + futures::Future<Output = ActorResult>>(
+    /// e.g. termination of the owning facet or crashing of the owning actor). Stops the active
+    /// facet when the linked task completes. Uses `name` for log messages emitted by the task.
+    pub fn linked_task<F: 'static + Send + futures::Future<Output = Result<LinkedTaskTermination, Error>>>(
         &mut self,
         name: tracing::Span,
         boot: F,
@@ -894,21 +905,18 @@ impl<'activation> Activation<'activation> {
                     let result = select! {
                         _ = token.cancelled() => {
                             tracing::trace!(task_id, "linked task cancelled");
-                            Ok(())
+                            LinkedTaskTermination::Normal
                         }
-                        result = boot => {
-                            match &result {
-                                Ok(()) => {
-                                    tracing::trace!(task_id, "linked task normal stop");
-                                    ()
-                                }
-                                Err(e) => {
-                                    tracing::error!(task_id, "linked task error: {}", e);
-                                    let _ = mailbox.tx.send(SystemMessage::Crash(e.clone()));
-                                    ()
-                                }
+                        result = boot => match result {
+                            Ok(t) => {
+                                tracing::trace!(task_id, "linked task normal stop");
+                                t
+                            }
+                            Err(e) => {
+                                tracing::error!(task_id, "linked task error: {}", e);
+                                let _ = mailbox.tx.send(SystemMessage::Crash(e.clone()));
+                                Err(e)?
                             }
-                            result
                         }
                     };
                     let _ = facet.activate(
@@ -917,9 +925,12 @@ impl<'activation> Activation<'activation> {
                                 tracing::trace!(task_id, "cancellation token removed");
                                 f.linked_tasks.remove(&task_id);
                             }
+                            if let LinkedTaskTermination::Normal = result {
+                                t.stop();
+                            }
                             Ok(())
                         });
-                    result
+                    Ok::<(), Error>(())
                 }.instrument(name));
             }
             f.linked_tasks.insert(task_id, token);
@@ -928,19 +939,30 @@ impl<'activation> Activation<'activation> {
 
     /// Executes the given action after the given duration has elapsed (so long as the active
     /// facet still exists at that time).
-    pub fn after(&mut self, duration: time::Duration, a: Action) {
+    pub fn after<F: 'static + Send + FnOnce(&mut Activation) -> ActorResult>(
+        &mut self,
+        duration: time::Duration,
+        a: F,
+    ) {
         self.at(time::Instant::now() + duration, a)
     }
 
     /// Executes the given action at the given instant (so long as the active facet still
     /// exists at that time).
-    pub fn at<I: Into<tokio::time::Instant>>(&mut self, instant: I, a: Action) {
-        let facet = self.facet.clone();
+    pub fn at<I: Into<tokio::time::Instant>, F: 'static + Send + FnOnce(&mut Activation) -> ActorResult>(
+        &mut self,
+        instant: I,
+        a: F,
+    ) {
         let account = Arc::clone(self.account());
         let instant = instant.into();
-        self.linked_task(crate::name!("Activation::at"), async move {
+        let facet = self.facet.clone();
+        let span = tracing::Span::current().clone();
+        self.linked_task(crate::name!(parent: None, "Activation::at"), async move {
             tokio::time::sleep_until(instant.into()).await;
-            facet.activate(account, a)
+            let _entry = span.enter();
+            facet.activate(account, a)?;
+            Ok(LinkedTaskTermination::KeepFacet)
         });
     }
 
@@ -987,7 +1009,8 @@ impl<'activation> Activation<'activation> {
     }
 
     /// Create a new subfacet of the currently-active facet. Runs `boot` in the new facet's
-    /// context. If `boot` returns leaving the new facet [inert][Facet#inert-facets],
+    /// context. If `boot` returns leaving the new facet [inert][Facet#inert-facets], the new
+    /// facet is [stopped][Activation::stop_facet].
     pub fn facet<F: FnOnce(&mut Activation) -> ActorResult>(
         &mut self,
         boot: F,
@@ -995,6 +1018,9 @@ impl<'activation> Activation<'activation> {
         let f = Facet::new(Some(self.facet.facet_id));
         let facet_id = f.facet_id;
         self.state.facet_nodes.insert(facet_id, f);
+        tracing::trace!(parent_id = ?self.facet.facet_id,
+                        ?facet_id,
+                        actor_facet_count = ?self.state.facet_nodes.len());
         self.state.facet_children.entry(self.facet.facet_id).or_default().insert(facet_id);
         self.with_facet(true /* TODO: tiny optimisation: "false" would be safe here */, facet_id, move |t| {
             boot(t)?;
@@ -1027,27 +1053,45 @@ impl<'activation> Activation<'activation> {
     }
 
     /// Arranges for the [`Facet`] named by `facet_id` to be stopped cleanly when `self`
-    /// commits. If `continuation` is supplied, the facet to be stopped hasn't been stopped
-    /// yet, none of the shutdown handlers yields an error, and the facet's parent facet is
-    /// alive, executes `continuation` in the parent facet's context.
-    pub fn stop_facet(&mut self, facet_id: FacetId, continuation: Option<Action>) {
+    /// commits.
+    ///
+    /// Then,
+    ///  - if `continuation` is supplied, and
+    ///  - the facet to be stopped hasn't been stopped at the time of the `stop_facet_and_continue` call, and
+    ///  - none of the shutdown handlers yields an error, and
+    ///  - the facet's parent facet is alive,
+    /// executes `continuation` (immediately) in the *parent* facet's context.
+    ///
+    pub fn stop_facet_and_continue<F: 'static + Send + FnOnce(&mut Activation) -> ActorResult>(
+        &mut self,
+        facet_id: FacetId,
+        continuation: Option<F>,
+    ) -> ActorResult {
         let maybe_parent_id = self.active_facet().and_then(|f| f.parent_facet_id);
-        self.enqueue_for_myself_at_commit(Box::new(move |t| {
-            t._terminate_facet(facet_id, true)?;
-            if let Some(k) = continuation {
-                if let Some(parent_id) = maybe_parent_id {
-                    t.with_facet(true, parent_id, k)?;
-                }
+        self.enqueue_for_myself_at_commit(Box::new(move |t| t._terminate_facet(facet_id, true)));
+        if let Some(k) = continuation {
+            if let Some(parent_id) = maybe_parent_id {
+                self.with_facet(true, parent_id, k)?;
             }
-            Ok(())
-        }));
+        }
+        Ok(())
+    }
+
+    /// Arranges for the [`Facet`] named by `facet_id` to be stopped cleanly when `self`
+    /// commits.
+    ///
+    /// Equivalent to `self.stop_facet_and_continue(facet_id, None)`, except that the lack of a
+    /// continuation means that there's no need for this method to return `ActorResult`.
+    pub fn stop_facet(&mut self, facet_id: FacetId) {
+        self.stop_facet_and_continue::<Action>(facet_id, None)
+            .expect("Non-failing stop_facet_and_continue")
     }
 
     /// Arranges for the active facet to be stopped cleanly when `self` commits.
     ///
-    /// Equivalent to `self.stop_facet(self.facet.facet_id, None)`.
+    /// Equivalent to `self.stop_facet(self.facet.facet_id)`.
     pub fn stop(&mut self) {
-        self.stop_facet(self.facet.facet_id, None)
+        self.stop_facet(self.facet.facet_id)
     }
 
     fn stop_if_inert(&mut self) {
@@ -1056,7 +1100,7 @@ impl<'activation> Activation<'activation> {
             tracing::trace!("Checking inertness of facet {} from facet {}", facet_id, t.facet.facet_id);
             if t.state.facet_exists_and_is_inert(facet_id) {
                 tracing::trace!(" - facet {} is inert, stopping it", facet_id);
-                t.stop_facet(facet_id, None);
+                t.stop_facet(facet_id);
             } else {
                 tracing::trace!(" - facet {} is not inert", facet_id);
             }
@@ -1066,7 +1110,8 @@ impl<'activation> Activation<'activation> {
 
     fn _terminate_facet(&mut self, facet_id: FacetId, alive: bool) -> ActorResult {
         if let Some(mut f) = self.state.facet_nodes.remove(&facet_id) {
-            tracing::debug!("{} termination of {:?}",
+            tracing::trace!(actor_facet_count = ?self.state.facet_nodes.len(),
+                            "{} termination of {:?}",
                             if alive { "living" } else { "post-exit" },
                             facet_id);
             if let Some(p) = f.parent_facet_id {
@@ -1088,9 +1133,13 @@ impl<'activation> Activation<'activation> {
                     // cleanup-actions are performed before parent-facet cleanup-actions.
                     if let Some(p) = parent_facet_id {
                         if t.state.facet_exists_and_is_inert(p) {
+                            tracing::trace!("terminating parent {:?} of facet {:?}", p, facet_id);
                             t._terminate_facet(p, true)?;
+                        } else {
+                            tracing::trace!("not terminating parent {:?} of facet {:?}", p, facet_id);
                         }
                     } else {
+                        tracing::trace!("terminating actor of root facet {:?}", facet_id);
                         t.state.shutdown();
                     }
                 } else {
@@ -1103,25 +1152,31 @@ impl<'activation> Activation<'activation> {
         }
     }
 
-    /// Create a new dataflow variable (field) within the active [`Actor`].
-    pub fn field<T: Any + Send>(&mut self, initial_value: T) -> Arc<Field<T>> {
+    /// Create a new named dataflow variable (field) within the active [`Actor`].
+    pub fn named_field<T: Any + Send>(&mut self, name: &str, initial_value: T) -> Arc<Field<T>> {
         let field_id = FieldId::new(NEXT_FIELD_ID.fetch_add(BUMP_AMOUNT.into(), Ordering::Relaxed))
             .expect("Internal error: Attempt to allocate FieldId of zero. Too many FieldIds allocated. Restart the process.");
         self.state.fields.insert(field_id, Box::new(initial_value));
         Arc::new(Field {
+            name: name.to_owned(),
             field_id,
             tx: self.state.tx.clone(),
             phantom: PhantomData,
         })
     }
 
+    /// Create a new anonymous dataflow variable (field) within the active [`Actor`].
+    pub fn field<T: Any + Send>(&mut self, initial_value: T) -> Arc<Field<T>> {
+        self.named_field("", initial_value)
+    }
+
     /// Retrieve a reference to the current value of a dataflow variable (field); if execution
     /// is currently within a [dataflow block][Activation::dataflow], marks the block as
     /// *depending upon* the field.
     ///
     pub fn get<T: Any + Send>(&mut self, field: &Field<T>) -> &T {
-        tracing::trace!(field = ?field.field_id, block = ?self.active_block, "get");
         if let Some(block) = self.active_block {
+            tracing::trace!(?field, ?block, action = "get", "observed");
             self.state.dataflow.record_observation(block, field.field_id);
         }
         let any = self.state.fields.get(&field.field_id)
@@ -1136,12 +1191,13 @@ impl<'activation> Activation<'activation> {
     /// reevaluation of dependent blocks.
     ///
     pub fn get_mut<T: Any + Send>(&mut self, field: &Field<T>) -> &mut T {
-        tracing::trace!(field = ?field.field_id, block = ?self.active_block, "get_mut");
         {
             // Overapproximation.
             if let Some(block) = self.active_block {
+                tracing::trace!(?field, ?block, action = "get_mut", "observed");
                 self.state.dataflow.record_observation(block, field.field_id);
             }
+            tracing::trace!(?field, active_block = ?self.active_block, action = "get_mut", "damaged");
             self.state.dataflow.record_damage(field.field_id);
         }
         let any = self.state.fields.get_mut(&field.field_id)
@@ -1153,7 +1209,7 @@ impl<'activation> Activation<'activation> {
     /// the new value is [`eq`][std::cmp::PartialEq::eq] to the value being overwritten.
     ///
     pub fn set<T: Any + Send>(&mut self, field: &Field<T>, value: T) {
-        tracing::trace!(field = ?field.field_id, block = ?self.active_block, "set");
+        tracing::trace!(?field, active_block = ?self.active_block, action = "set", "damaged");
         // Overapproximation in many cases, since the new value may not produce an
         // observable difference (may be equal to the current value).
         self.state.dataflow.record_damage(field.field_id);
@@ -1200,7 +1256,9 @@ impl<'activation> Activation<'activation> {
                 }
             }
         }
-        tracing::trace!(passes = ?pass_number, "repair_dataflow complete");
+        if pass_number > 0 {
+            tracing::trace!(passes = ?pass_number, "repair_dataflow complete");
+        }
         Ok(pass_number > 0)
     }
 
@@ -1291,7 +1349,7 @@ impl Account {
     pub fn new(name: tracing::Span) -> Arc<Self> {
         let id = NEXT_ACCOUNT_ID.fetch_add(1, Ordering::Relaxed);
         let debt = Arc::new(AtomicI64::new(0));
-        ACCOUNTS.write().unwrap().insert(id, (name, Arc::clone(&debt)));
+        ACCOUNTS.write().insert(id, (name, Arc::clone(&debt)));
         Arc::new(Account {
             id,
             debt,
@@ -1334,7 +1392,7 @@ impl Account {
 
 impl Drop for Account {
     fn drop(&mut self) {
-        ACCOUNTS.write().unwrap().remove(&self.id);
+        ACCOUNTS.write().remove(&self.id);
     }
 }
 
@@ -1397,6 +1455,7 @@ impl PartialOrd for Mailbox {
 
 impl Drop for Mailbox {
     fn drop(&mut self) {
+        tracing::debug!("Last reference to mailbox of actor id {:?} was dropped", self.actor_id);
         let _ = self.tx.send(SystemMessage::Release);
         ()
     }
@@ -1408,7 +1467,7 @@ impl Actor {
         let (tx, rx) = unbounded_channel();
         let actor_id = next_actor_id();
         let root = Facet::new(None);
-        // tracing::trace!(id = actor_id, "Actor::new");
+        tracing::trace!(?actor_id, root_facet_id = ?root.facet_id, "Actor::new");
         let mut st = RunningActor {
             actor_id,
             tx,
@@ -1423,13 +1482,8 @@ impl Actor {
             root: root.facet_id,
         };
         st.facet_nodes.insert(root.facet_id, root);
-        Actor {
-            rx,
-            ac_ref: ActorRef {
-                actor_id,
-                state: Arc::new(Mutex::new(ActorState::Running(st))),
-            },
-        }
+        let ac_ref = ActorRef { actor_id, state: Arc::new(Mutex::new(ActorState::Running(st))) };
+        Actor { rx, ac_ref }
     }
 
     fn link(self, t_parent: &mut Activation) -> Self {
@@ -1453,6 +1507,7 @@ impl Actor {
         name: tracing::Span,
         boot: F,
     ) -> ActorHandle {
+        ACTORS.write().insert(self.ac_ref.actor_id, (name.clone(), self.ac_ref.clone()));
         name.record("actor_id", &self.ac_ref.actor_id);
         tokio::spawn(async move {
             tracing::trace!("start");
@@ -1498,7 +1553,7 @@ impl Actor {
                     SystemMessage::ReleaseField(field_id) => {
                         tracing::trace!(actor_id = ?self.ac_ref.actor_id,
                                         "SystemMessage::ReleaseField({})", field_id);
-                        self.ac_ref.access(|s| if let ActorState::Running(ra) = s.unwrap() {
+                        self.ac_ref.access(|s| if let ActorState::Running(ra) = s {
                             ra.fields.remove(&field_id);
                         })
                     }
@@ -1557,19 +1612,12 @@ impl Facet {
     }
 }
 
-fn panicked_err() -> ActorResult {
-    Err(error("Actor panicked", AnyValue::new(false)))
-}
-
 impl ActorRef {
     /// Uses an internal mutex to access the internal state: takes the
     /// lock, calls `f` with the internal state, releases the lock,
     /// and returns the result of `f`.
-    pub fn access<R, F: FnOnce(Option<&mut ActorState>) -> R>(&self, f: F) -> R {
-        match self.state.lock() {
-            Err(_) => f(None),
-            Ok(mut g) => f(Some(&mut *g)),
-        }
+    pub fn access<R, F: FnOnce(&mut ActorState) -> R>(&self, f: F) -> R {
+        f(&mut *self.state.lock())
     }
 
     /// Retrieves the exit status of the denoted actor. If it is still
@@ -1577,12 +1625,10 @@ impl ActorRef {
     /// exited normally, or `Some(Err(_))` if it terminated
     /// abnormally.
     pub fn exit_status(&self) -> Option<ActorResult> {
-        self.access(|s| s.map_or_else(
-            || Some(panicked_err()),
-            |state| match state {
-                ActorState::Running(_) => None,
-                ActorState::Terminated { exit_status } => Some((**exit_status).clone()),
-            }))
+        self.access(|state| match state {
+            ActorState::Running(_) => None,
+            ActorState::Terminated { exit_status } => Some((**exit_status).clone()),
+        })
     }
 
     fn facet_ref(&self, facet_id: FacetId) -> FacetRef {
@@ -1593,7 +1639,7 @@ impl ActorRef {
     }
 
     fn root_facet_id(&self) -> FacetId {
-        self.access(|s| match s.expect("Actor missing its state") {
+        self.access(|s| match s {
             ActorState::Terminated { .. } => panic!("Actor unexpectedly in terminated state"),
             ActorState::Running(ra) => ra.root, // what a lot of work to get this one number
         })
@@ -1604,6 +1650,12 @@ impl ActorRef {
     }
 }
 
+impl std::fmt::Debug for ActorRef {
+    fn fmt(&self, f: &mut std::fmt::Formatter) -> Result<(), std::fmt::Error> {
+        write!(f, "#<ActorRef {}>", self.actor_id)
+    }
+}
+
 impl RunningActor {
     /// Requests a shutdown of the actor. The shutdown request is
     /// handled by the actor's main loop, causing it to terminate with
@@ -1643,11 +1695,17 @@ impl RunningActor {
     fn facet_exists_and_is_inert(&mut self, facet_id: FacetId) -> bool {
         let no_kids = self.facet_children.get(&facet_id).map(|cs| cs.is_empty()).unwrap_or(true);
         if let Some(f) = self.get_facet(facet_id) {
-            no_kids &&
-                f.outbound_handles.is_empty() &&
-                f.linked_tasks.is_empty() &&
-                f.inert_check_preventers.load(Ordering::Relaxed) == 0
+            // The only outbound handle the root facet of an actor may have is a link
+            // assertion, from [Activation::link]. This is not to be considered a "real"
+            // assertion for purposes of keeping the facet alive!
+            let no_outbound_handles = f.outbound_handles.is_empty();
+            let is_root_facet = f.parent_facet_id.is_none();
+            let no_linked_tasks = f.linked_tasks.is_empty();
+            let no_inert_check_preventers = f.inert_check_preventers.load(Ordering::Relaxed) == 0;
+            tracing::trace!(?facet_id, ?no_kids, ?no_outbound_handles, ?is_root_facet, ?no_linked_tasks, ?no_inert_check_preventers);
+            no_kids && (no_outbound_handles || is_root_facet) && no_linked_tasks && no_inert_check_preventers
         } else {
+            tracing::trace!(?facet_id, exists = ?false);
             false
         }
     }
@@ -1671,6 +1729,12 @@ impl RunningActor {
     }
 }
 
+impl<T: Any + Send> std::fmt::Debug for Field<T> {
+    fn fmt(&self, f: &mut std::fmt::Formatter) -> Result<(), std::fmt::Error> {
+        write!(f, "#<Field {:?} {}>", self.name, self.field_id)
+    }
+}
+
 impl<T: Any + Send> Eq for Field<T> {}
 impl<T: Any + Send> PartialEq for Field<T> {
     fn eq(&self, other: &Field<T>) -> bool {
@@ -1688,6 +1752,7 @@ impl<T: Any + Send> Drop for Field<T> {
 impl Drop for Actor {
     fn drop(&mut self) {
         self.rx.close();
+        ACTORS.write().remove(&self.ac_ref.actor_id);
         tracing::trace!(actor_id = ?self.ac_ref.actor_id, "Actor::drop");
     }
 }
@@ -1730,7 +1795,7 @@ impl<M> Ref<M> {
     ///
     /// Panics if this `Ref` has already been given a behaviour.
     pub fn become_entity<E: 'static + Entity<M>>(&self, e: E) {
-        let mut g = self.target.lock().expect("unpoisoned");
+        let mut g = self.target.lock();
         if g.is_some() {
             panic!("Double initialization of Ref");
         }
@@ -1738,7 +1803,7 @@ impl<M> Ref<M> {
     }
 
     fn internal_with_entity<R, F: FnOnce(&mut dyn Entity<M>) -> R>(&self, f: F) -> R {
-        let mut g = self.target.lock().expect("unpoisoned");
+        let mut g = self.target.lock();
         f(g.as_mut().expect("initialized").as_mut())
     }
 }
@@ -1861,6 +1926,11 @@ impl Cap {
             t.message(&self.underlying, m)
         }
     }
+
+    /// Synchronizes with the reference underlying the cap.
+    pub fn sync(&self, t: &mut Activation, peer: Arc<Ref<Synced>>) {
+        t.sync(&self.underlying, peer)
+    }
 }
 
 impl std::fmt::Debug for Cap {
@@ -1925,8 +1995,7 @@ where
 
 impl<M> Entity<M> for StopOnRetract {
     fn retract(&mut self, t: &mut Activation, _h: Handle) -> ActorResult {
-        t.stop();
-        Ok(())
+        Ok(t.stop())
     }
 }
 
@@ -1936,6 +2005,47 @@ impl<F: Send + FnMut(&mut Activation) -> ActorResult> Entity<Synced> for F {
     }
 }
 
+async fn wait_loop(wait_time: time::Duration) {
+    let deadline = time::Instant::now() + wait_time;
+    while time::Instant::now() < deadline {
+        let remaining_count = ACTORS.read().len();
+        if remaining_count == 0 {
+            break;
+        }
+        tracing::debug!("Waiting for {} remaining actors to stop", remaining_count);
+        tokio::time::sleep(time::Duration::from_millis(100)).await;
+    }
+}
+
+pub async fn wait_for_all_actors_to_stop(wait_time: time::Duration) {
+    wait_loop(wait_time).await;
+    let remaining = ACTORS.read().clone();
+    if remaining.len() > 0 {
+        tracing::warn!("Some actors remain after {:?}:", wait_time);
+        for (name, actor) in remaining.into_values() {
+            let _entry = name.enter();
+            tracing::warn!(?actor, "still running, requesting shutdown");
+            let g = actor.state.lock();
+            if let ActorState::Running(state) = &*g {
+                state.shutdown();
+            }
+        }
+        wait_loop(wait_time).await;
+        let remaining = ACTORS.read().clone();
+        if remaining.len() > 0 {
+            tracing::error!("Some actors failed to stop after being explicitly shut down:");
+            for (name, actor) in remaining.into_values() {
+                let _entry = name.enter();
+                tracing::error!(?actor, "failed to stop");
+            }
+        } else {
+            tracing::debug!("All remaining actors have stopped.");
+        }
+    } else {
+        tracing::debug!("All remaining actors have stopped.");
+    }
+}
+
 /// A convenient Syndicate-enhanced variation on
 /// [`tracing::info_span`].
 ///
