@@ -15,6 +15,8 @@ use futures::SinkExt;
 use futures::Stream;
 use futures::StreamExt;
 
+use parking_lot::Mutex;
+
 use preserves::error::Error as PreservesError;
 use preserves::error::is_eof_io_error;
 use preserves::value::BinarySource;
@@ -37,7 +39,6 @@ use preserves_schema::ParseError;
 use std::io;
 use std::pin::Pin;
 use std::sync::Arc;
-use std::sync::Mutex;
 use std::sync::atomic::AtomicUsize;
 use std::sync::atomic::Ordering;
 
@@ -177,6 +178,7 @@ pub fn connect_stream<I, O, E, F>(
     t: &mut Activation,
     i: I,
     o: O,
+    output_text: bool,
     sturdyref: sturdy::SturdyRef,
     initial_state: E,
     mut f: F,
@@ -188,7 +190,7 @@ pub fn connect_stream<I, O, E, F>(
 {
     let i = Input::Bytes(Box::pin(i));
     let o = Output::Bytes(Box::pin(o));
-    let gatekeeper = TunnelRelay::run(t, i, o, None, Some(sturdy::Oid(0.into()))).unwrap();
+    let gatekeeper = TunnelRelay::run(t, i, o, None, Some(sturdy::Oid(0.into())), output_text).unwrap();
     let main_entity = t.create(during::entity(initial_state).on_asserted(move |state, t, a: AnyValue| {
         let denotation = a.value().to_embedded()?;
         f(state, t, Arc::clone(denotation))
@@ -218,6 +220,7 @@ impl TunnelRelay {
         o: Output,
         initial_ref: Option<Arc<Cap>>,
         initial_oid: Option<sturdy::Oid>,
+        output_text: bool,
     ) -> Option<Arc<Cap>> {
         let (output_tx, output_rx) = unbounded_channel();
         let tr_ref = Arc::new(Mutex::new(None));
@@ -227,7 +230,7 @@ impl TunnelRelay {
         let mut tr = TunnelRelay {
             self_ref: Arc::clone(&tr_ref),
             output: output_tx,
-            output_text: false,
+            output_text,
             inbound_assertions: Map::new(),
             outbound_assertions: Map::new(),
             membranes: Membranes {
@@ -244,7 +247,7 @@ impl TunnelRelay {
         let result = initial_oid.map(
             |io| Arc::clone(&tr.membranes.import_oid(t, &tr_ref, io).inc_ref().obj));
         dump_membranes!(tr.membranes);
-        *tr_ref.lock().unwrap() = Some(tr);
+        *tr_ref.lock() = Some(tr);
         t.linked_task(crate::name!("writer"), output_loop(o, output_rx));
         t.linked_task(crate::name!("reader"), input_loop(t.facet.clone(), i, tr_ref));
         t.state.add_exit_hook(&self_entity);
@@ -276,12 +279,14 @@ impl TunnelRelay {
     }
 
     fn handle_inbound_datagram(&mut self, t: &mut Activation, bs: &[u8]) -> ActorResult {
+        tracing::trace!(bytes = ?bs, "inbound datagram");
         let item = self.deserialize_one(t, bs).0?;
         self.handle_inbound_packet(t, item)
     }
 
     fn handle_inbound_stream(&mut self, t: &mut Activation, buf: &mut BytesMut) -> ActorResult {
         loop {
+            tracing::trace!(buffer = ?buf, "inbound stream");
             let (result, count) = self.deserialize_one(t, buf);
             match result {
                 Err(ParseError::Preserves(PreservesError::Io(e)))
@@ -296,7 +301,7 @@ impl TunnelRelay {
     }
 
     fn handle_inbound_packet(&mut self, t: &mut Activation, p: P::Packet<AnyValue>) -> ActorResult {
-        tracing::trace!(packet = ?p, "-->");
+        tracing::debug!(packet = ?p, "-->");
         match p {
             P::Packet::Error(b) => {
                 tracing::info!(message = ?b.message.clone(),
@@ -311,7 +316,7 @@ impl TunnelRelay {
                         Some(ws) =>
                             ws.inc_ref(),
                         None => {
-                            tracing::warn!(
+                            tracing::debug!(
                                 event = ?language().unparse(&P::TurnEvent { oid, event }),
                                 "Cannot deliver event: nonexistent oid");
                             return Ok(());
@@ -371,7 +376,7 @@ impl TunnelRelay {
                             impl Entity<Synced> for SyncPeer {
                                 fn message(&mut self, t: &mut Activation, _a: Synced) -> ActorResult {
                                     self.peer.message(t, &(), &AnyValue::new(true));
-                                    let mut g = self.relay_ref.lock().expect("unpoisoned");
+                                    let mut g = self.relay_ref.lock();
                                     let tr = g.as_mut().expect("initialized");
                                     tr.membranes.release(std::mem::take(&mut self.pins));
                                     dump_membranes!(tr.membranes);
@@ -438,7 +443,7 @@ impl TunnelRelay {
 
     pub fn send_packet(&mut self, account: &Arc<Account>, cost: usize, p: P::Packet<AnyValue>) -> ActorResult {
         let item = language().unparse(&p);
-        tracing::trace!(packet = ?item, "<--");
+        tracing::debug!(packet = ?item, "<--");
 
         let bs = if self.output_text {
             let mut s = TextWriter::encode(&mut self.membranes, &item)?;
@@ -598,16 +603,16 @@ async fn input_loop(
     facet: FacetRef,
     i: Input,
     relay: TunnelRelayRef,
-) -> ActorResult {
+) -> Result<LinkedTaskTermination, Error> {
     let account = Account::new(crate::name!("input-loop"));
     match i {
         Input::Packets(mut src) => {
             loop {
                 account.ensure_clear_funds().await;
                 match src.next().await {
-                    None => return facet.activate(Arc::clone(&account), |t| Ok(t.state.shutdown())),
+                    None => return Ok(LinkedTaskTermination::Normal),
                     Some(bs) => facet.activate(Arc::clone(&account), |t| {
-                        let mut g = relay.lock().expect("unpoisoned");
+                        let mut g = relay.lock();
                         let tr = g.as_mut().expect("initialized");
                         tr.handle_inbound_datagram(t, &bs?)
                     })?,
@@ -624,15 +629,15 @@ async fn input_loop(
                     Ok(n) => n,
                     Err(e) =>
                         if e.kind() == io::ErrorKind::ConnectionReset {
-                            return facet.activate(Arc::clone(&account), |t| Ok(t.state.shutdown()));
+                            return Ok(LinkedTaskTermination::Normal);
                         } else {
                             return Err(e)?;
                         },
                 };
                 match n {
-                    0 => return facet.activate(Arc::clone(&account), |t| Ok(t.state.shutdown())),
+                    0 => return Ok(LinkedTaskTermination::Normal),
                     _ => facet.activate(Arc::clone(&account), |t| {
-                        let mut g = relay.lock().expect("unpoisoned");
+                        let mut g = relay.lock();
                         let tr = g.as_mut().expect("initialized");
                         tr.handle_inbound_stream(t, &mut buf)
                     })?,
@@ -645,11 +650,11 @@ async fn input_loop(
 async fn output_loop(
     mut o: Output,
     mut output_rx: UnboundedReceiver<LoanedItem<Vec<u8>>>,
-) -> ActorResult {
+) -> Result<LinkedTaskTermination, Error> {
     loop {
         match output_rx.recv().await {
             None =>
-                return Ok(()),
+                return Ok(LinkedTaskTermination::KeepFacet),
             Some(mut loaned_item) => {
                 match &mut o {
                     Output::Packets(sink) => sink.send(std::mem::take(&mut loaned_item.item)).await?,
@@ -665,7 +670,7 @@ async fn output_loop(
 
 impl Entity<()> for TunnelRefEntity {
     fn message(&mut self, t: &mut Activation, _m: ()) -> ActorResult {
-        let mut g = self.relay_ref.lock().expect("unpoisoned");
+        let mut g = self.relay_ref.lock();
         let tr = g.as_mut().expect("initialized");
         let events = std::mem::take(&mut tr.pending_outbound);
         tr.send_packet(&t.account(), events.len(), P::Packet::Turn(Box::new(P::Turn(events.clone()))))?;
@@ -678,7 +683,7 @@ impl Entity<()> for TunnelRefEntity {
     fn exit_hook(&mut self, t: &mut Activation, exit_status: &Arc<ActorResult>) -> ActorResult {
         if let Err(e) = &**exit_status {
             let e = e.clone();
-            let mut g = self.relay_ref.lock().expect("unpoisoned");
+            let mut g = self.relay_ref.lock();
             let tr = g.as_mut().expect("initialized");
             tr.send_packet(&t.account(), 1, P::Packet::Error(Box::new(e)))?;
         }
@@ -688,32 +693,28 @@ impl Entity<()> for TunnelRefEntity {
 
 impl Entity<AnyValue> for RelayEntity {
     fn assert(&mut self, t: &mut Activation, a: AnyValue, h: Handle) -> ActorResult {
-        let mut g = self.relay_ref.lock().expect("unpoisoned");
-        let tr = g.as_mut().expect("initialized");
-        tr.send_event(t, self.oid.clone(), P::Event::Assert(Box::new(P::Assert {
-            assertion: P::Assertion(a),
-            handle: P::Handle(h.into()),
-        })))
+        self.relay_ref.lock().as_mut().expect("initialized")
+            .send_event(t, self.oid.clone(), P::Event::Assert(Box::new(P::Assert {
+                assertion: P::Assertion(a),
+                handle: P::Handle(h.into()),
+            })))
     }
     fn retract(&mut self, t: &mut Activation, h: Handle) -> ActorResult {
-        let mut g = self.relay_ref.lock().expect("unpoisoned");
-        let tr = g.as_mut().expect("initialized");
-        tr.send_event(t, self.oid.clone(), P::Event::Retract(Box::new(P::Retract {
-            handle: P::Handle(h.into()),
-        })))
+        self.relay_ref.lock().as_mut().expect("initialized")
+            .send_event(t, self.oid.clone(), P::Event::Retract(Box::new(P::Retract {
+                handle: P::Handle(h.into()),
+            })))
     }
     fn message(&mut self, t: &mut Activation, m: AnyValue) -> ActorResult {
-        let mut g = self.relay_ref.lock().expect("unpoisoned");
-        let tr = g.as_mut().expect("initialized");
-        tr.send_event(t, self.oid.clone(), P::Event::Message(Box::new(P::Message {
-            body: P::Assertion(m)
-        })))
+        self.relay_ref.lock().as_mut().expect("initialized")
+            .send_event(t, self.oid.clone(), P::Event::Message(Box::new(P::Message {
+                body: P::Assertion(m)
+            })))
     }
     fn sync(&mut self, t: &mut Activation, peer: Arc<Ref<Synced>>) -> ActorResult {
-        let mut g = self.relay_ref.lock().expect("unpoisoned");
-        let tr = g.as_mut().expect("initialized");
-        tr.send_event(t, self.oid.clone(), P::Event::Sync(Box::new(P::Sync {
-            peer: Cap::guard(Arc::new(()), peer)
-        })))
+        self.relay_ref.lock().as_mut().expect("initialized")
+            .send_event(t, self.oid.clone(), P::Event::Sync(Box::new(P::Sync {
+                peer: Cap::guard(Arc::new(()), peer)
+            })))
     }
 }
